#include <iostream>

using namespace std;

#define AUSGABE(x) cout << (x) << endl;




// noch mehr zum Thema Zuweisungsoperator

class DTest // reserviert Speicher in whlbarer Gre
{
    private:
    int length; // nach der Initialisierung: Gre des reservierten Speichers
    int* p; // nach der Initialisierung: Zeiger auf den reservierten Speicher

    public:

    DTest(int l) :
    length(l)
    {
        AUSGABE("Konstruktor")
        p = new int[length]; // Speicher reservieren
    }

    DTest(const DTest& other)
    {
        AUSGABE("Copy-Konstruktor")
        p = new int[other.length]; // Speicher reservieren
        length = other.length;
        for (int i=0; i<length; i++)
        {
            p[i] = (other.p)[i]; // tiefe Kopie
        }
    }

    ~DTest() // Wir brauchen einen Destruktor, ...
    {
        delete[] p; // ... um den Speicher freizugeben.
    }

    // Zuweisungsoperator fr Zusammenspiel mit STL (kommt spter)
    // nur ohne const fr Ergebnis. Aber dann kann man gefhrliche
    // Ausdrcke wie (A=B)=C bilden.
    const DTest& operator= (const DTest& other)
    {
        AUSGABE("Zuweisungsoperator")
        // Mssen fr tiefe Kopie Speicherinhalt des Objekts other
        // in unser eigenes Objekt *this bringen.
        // Problem: Gre des jeweils reservierten Speichers ist ggf. verschieden.

        int* pNew = new int[other.length]; // genug Speicher reservieren
        for (int i=0; i<other.length; i++) // dort hinein kopieren
        {
            pNew[i] = (other.p)[i];
        }

        delete[] p; // alten Speicher freigeben

        p = pNew;
        length = other.length;

        return *this;
    }

    // Oft findet man folgende Lsung. Die ist nicht Exception-sicher!
/*
    const DTest& operator= (const DTest& other)
    {
        AUSGABE("Zuweisungsoperator")
        if (this != &other) // Test auf Selbstzuweisung.
                            // Ohne diesen Test funktioniert diese Lsung nicht,
                            // ein klarer Hinweis auf Exception-Probleme.
        {
            delete[] p; // Wenn in den folgenden Zeilen etwas passiert (new!),
                        // sind unsere alten Daten futsch. Schlecht!

            length = other.length;
            p = new int[length];
            for (int i=0; i<length; i++) // kopieren
            {
                p[i] = (other.p)[i];
            }
        }
        return *this;
    }
*/
    int getAt (int position) const
    {
        return p[position];
    }

    void setAt (int position, int value)
    {
        p[position] = value;
    }
};

DTest test (DTest a)
{
    return a;
}

int main (int argc, char* (argv[]))
{
    DTest a(13);
    DTest b(42);
    AUSGABE("-----------------------------------------")
    a = b;
    AUSGABE("-----------------------------------------")
    a = DTest(42);
    AUSGABE("-----------------------------------------")
    DTest c = a;
    AUSGABE("-----------------------------------------")
    DTest d(a);
    AUSGABE("-----------------------------------------")
    a = test(b);
    return 0;
}
