#include <stdexcept>
#include <new>
#include <iostream>
#include <string>

using namespace std;

#define AUSGABE(x) cout << (x) << endl;

// Thema: Exceptions

// Exceptions drfen in C++ anders als in Java beliebige interne Typen oder Objekte sein:
// throw 42;                     --->    catch (int e) { /*...*/ }
// throw "Fehler 42";            --->    catch (const char* e) { /*...*/ }
// throw string("Fehler 42");    --->    catch (string e) { /*...*/ }
// Am besten verwendet man aber die Exceptions aus #include<stdexcept>
// oder leitet von ihnen oder ihrer Mutterklasse exception ab. Ergebnis hnelt Java.

/* in stdexcept definierte Exceptions:

exception --+--> logic_error        fr logische Fehler
            `--> runtime_error      fr "echte" Laufzeitfehler

logic_error --+--> domain_error     Vorbedingung verletzt
              |--> length_error     Objekt zu lang
              |--> out_of_range     auerhalb Bereich
              `--> ...

runtime_error --+--> range_error    Nachbedigung verletzt
                |--> overflow_error berlauf
                `--> bad_alloc      Speicherreservierung fehlgeschlagen
*/

class NoSuchDate:    // unsere eigene Exception fr unmgliches Datum
public logic_error
{
    public:
    NoSuchDate(const string& what_arg) :     // Konstruktor berladen,
    logic_error (what_arg)                   // damit als NoSuchDate("Methode 42") aufrufbar.
    { }                                      // Mit NoSuchDate.what() kann man den String
};                                           // beim Behandeln der Exception erfahren.

// Eine Methode/Funktion ohne Deklaration throw( /*...*/ ) darf jede Exception werfen.
// Eine Methode/Funktion mit Dekaration throw( /*...*/ ) darf nur die angegebenen Exceptions werfen,
// sonst ruft C++ die globale Funktion unexpected() auf.
// Vorsicht: throw, nicht throws wie in Java
// C++ prft nicht wie Java, welche Exceptions auftreten knnten und ob sie behandelt werden.

int daysSince1900(int year, int month, int day) throw (NoSuchDate)
{
    int n = 42;
    // grere Rechnung
    // darin z.B.:
    if (year == 2001 && month == 2 && day == 29)
    {
        throw NoSuchDate("invalid input 2001-02-29 to daysSince1900");
    }
    // oder so etwas (ooops...)
    if (year == 2001 && month == 2 && day == 31)
    {
        int k = 50000;
        int l = 10000;
        int* a = new int[k*l]; // Problem: Aufruf wirft bad_alloc,
                               // das ist aber "unexpected" mangels Deklaration.
        a[13] = 42;
    }

    return n;
}



void myUnexpected()
{
    AUSGABE("unexpected exception")
    exit(42); // Programm mit Fehlerwert 42 beenden.
    // Die Unexpected-Funktion kann nicht (direkt) zurckkehren!
}



int main (int argc, char* (argv[]))
{
    set_unexpected(myUnexpected); // die eigene Unexpected-Funktion installieren

    for (int i = 1; i <= 31; i++)
    {
        try
        {
            daysSince1900(2001, 2, i);
            AUSGABE(i);
        }
        catch (NoSuchDate e)
        {
            AUSGABE(e.what())
        }
    }
    return 0;
}



// Jetzt die Klasse DTest mit Exceptions:

class DTest
{
    private:
    int length;
    int* p;

    public:

    ~DTest() throw() // Destruktor darf keine Exceptions werfen,
                     // sonst kann man fr nichts garantieren.
    {
        delete[] p;  // Standardmiges delete wirft keine Exceptions.
    }

    DTest(int l) throw (bad_alloc) : // fehlschlagendes new wirft bad_alloc, wird nicht gefangen
    length(l)
    {
        p = new int[length];
    }

    DTest(const DTest& other) throw (bad_alloc) // fehlschlagendes new wirft bad_alloc, wird nicht gefangen
    {
        p = new int[other.length];
        length = other.length;
        for (int i=0; i<length; i++)
        {
            p[i] = (other.p)[i];
        }
    }

    DTest& operator= (const DTest& other) throw (bad_alloc) // fehlschlagendes new wirft bad_alloc, wird nicht gefangen
    {
        int* pNew = new int[other.length];
        for (int i=0; i<other.length; i++)
        {
            pNew[i] = (other.p)[i];
        }

        delete[] p;

        p = pNew;
        length = other.length;

        return *this;
    }

    int getAt (int position) const throw (out_of_range)
    {
        if (position >= length || position < 0) // oder assert(/*...*/) verwenden, kommt spter
        {
            throw out_of_range("getAt with wrong argument");
        }

        return p[position];
    }

    void setAt (int position, int value) throw (out_of_range)
    {
        if (position >= length || position < 0)
        {
            throw out_of_range("setAt with wrong argument");
        }

        p[position] = value;
    }
};
